# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base logging configuration parameters."""

from pydantic import BaseModel, Field


class BaseLoggingParams(BaseModel):
    """Base configuration parameters for logging in interceptors."""

    log_level: str = Field(
        default="INFO",
        description="Log level for this interceptor. Can be overridden by NEMO_EVALUATOR_LOG_LEVEL environment variable.",
    )
